<?php

namespace Drupal\commerce_merchant_warrior\Api;

use Psr\Log\LoggerInterface;
use Drupal\Core\Config\Config;
use GuzzleHttp\ClientInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Config\ConfigFactoryInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\commerce_payment\Exception\HardDeclineException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;

/**
 * Service for implementation of the API for MerchantWarrior.
 */
class MerchantWarriorApi implements MerchantWarriorApiInterface {

  use StringTranslationTrait;
  use LoggerChannelTrait;

  /**
   * The Payframe config.
   *
   * @var \Drupal\Core\Config\Config
   */
  protected Config $config;

  /**
   * The HTTP client to fetch the access token.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected ClientInterface $httpClient;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected RequestStack $requestStack;

  /**
   * The Merchant Warrior request headers.
   */
  const HEADERS = [
    'Accept' => 'application/json',
    'Content-Type' => 'application/x-www-form-urlencoded',
  ];

  /**
   * Creates a new MerchantWarriorApi object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The Guzzle HTTP client.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    ClientInterface $http_client,
    RequestStack $request_stack
  ) {
    // Load the Merchant Warrior payment gateway configuration.
    // @todo Figure out how to inject the config directly.
    $this->config = $config_factory->get('commerce_payment.commerce_payment_gateway.merchant_warrior_payframe');
    $this->httpClient = $http_client;
    $this->logger = $this->getLogger('commerce_merchant_warrior');
    $this->requestStack = $request_stack;
  }

  /**
   * {@inheritdoc}
   */
  public function getAccessToken(): string {
    $config = $this->config->get('configuration');
    $merchant_uuid = $config['merchant_uuid'];
    $merchant_api_key = $config['api_key'];

    try {
      $options = [
        'form_params' => [
          'method' => 'getAccessToken',
          'merchantUUID' => $merchant_uuid,
          'apiKey' => $merchant_api_key,
        ],
        'headers' => self::HEADERS,
      ];
      $response = $this->httpClient->request('POST', $this->getApiUrl(), $options);
      $data = $response->getBody()->getContents();
      $encode_data = json_encode(simplexml_load_string($data));
      $decoded_data = json_decode($encode_data, TRUE);

      // Return the token if we have a successfuly response.
      if ($decoded_data['responseCode'] === '0') {
        return $decoded_data['token'];
      }
      else {
        $this->logger->error('Merchant Warrior getAccessToken(): Unable to fetch an access token. The error response: %message', [
          '%message' => $decoded_data['responseMessage'],
        ]);
        throw new PaymentGatewayException(
          sprintf('Merchant Warrior: Unable to fetch an access token from Merchant Warrior. The error response was: %s.',
          $decoded_data['responseMessage'],
        ));
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Merchant Warrior getAccessToken(): Unable to fetch an access token. The error response: %message', [
        '%message' => $e->getMessage(),
      ]);
      throw new PaymentGatewayException($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function processAuthorization(array $transaction_data): array {
    $config = $this->config->get('configuration');
    $merchant_uuid = $config['merchant_uuid'];
    $merchant_api_key = $config['api_key'];
    $merchant_api_passphrase = $config['api_passphrase'];

    // Generate the transaction hash.
    $hash = $this->generateHash([
      md5($merchant_api_passphrase),
      $merchant_uuid,
      $transaction_data['transactionAmount'],
      $transaction_data['transactionCurrency'],
    ]);

    try {
      $options = [
        'form_params' => [
          'method' => 'processAuth',
          'merchantUUID' => $merchant_uuid,
          'apiKey' => $merchant_api_key,
          'hash' => urlencode($hash),
        ],
        'headers' => self::HEADERS,
      ];
      // Append the transaction data received to the options form_params.
      $options['form_params'] = $options['form_params'] + $transaction_data;
      $response = $this->httpClient->request('POST', $this->getApiUrl(), $options);
      $data = $response->getBody()->getContents();
      $encode_data = json_encode(simplexml_load_string($data));
      $decoded_data = json_decode($encode_data, TRUE);

      // Return the data if we have a successfuly response.
      if ($decoded_data['responseCode'] === '0') {
        return $decoded_data;
      }
      else {
        $this->logger->error('Merchant Warrior processAuthorization(): Unable to process authorization for Order ID: %s. The error response: %message', [
          '%s' => $transaction_data['order_id'],
          '%message' => $decoded_data['responseMessage'],
        ]);
        throw new HardDeclineException(sprintf('Merchant Warrior: Unable to authorize the payment for Order ID: %s. The error response was: %s.',
          $transaction_data['order_id'],
          $decoded_data['responseMessage'],
        ));
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Merchant Warrior processAuthorization(): Unable to process authorization for Order ID: %s. The error response: %message', [
        '%s' => $transaction_data['order_id'],
        '%message' => $e->getMessage(),
      ]);
      throw new PaymentGatewayException($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function processTokenAuthorization(array $transaction_data): array {
    $config = $this->config->get('configuration');
    $merchant_uuid = $config['merchant_uuid'];
    $merchant_api_key = $config['api_key'];
    $merchant_api_passphrase = $config['api_passphrase'];

    // Generate the transaction hash.
    $hash = $this->generateHash([
      md5($merchant_api_passphrase),
      $merchant_uuid,
      $transaction_data['transactionAmount'],
      $transaction_data['transactionCurrency'],
    ]);

    try {
      $options = [
        'form_params' => [
          'method' => 'processAuth',
          'merchantUUID' => $merchant_uuid,
          'apiKey' => $merchant_api_key,
          'hash' => urlencode($hash),
        ],
        'headers' => self::HEADERS,
      ];
      // Append the transaction data received to the options form_params.
      $options['form_params'] = $options['form_params'] + $transaction_data;
      $response = $this->httpClient->request('POST', $this->getTokenApiUrl(), $options);
      $data = $response->getBody()->getContents();
      $encode_data = json_encode(simplexml_load_string($data));
      $decoded_data = json_decode($encode_data, TRUE);

      // Return the data if we have a successfuly response.
      if ($decoded_data['responseCode'] === '0') {
        return $decoded_data;
      }
      else {
        $this->logger->error('Merchant Warrior processTokenAuthorization(): Unable to authorize the access token for Order ID: %s. The error response: %message', [
          '%s' => $transaction_data['order_id'],
          '%message' => $decoded_data['responseMessage'],
        ]);
        throw new HardDeclineException(sprintf('Merchant Warrior: Unable to authorize the payment for Order ID: %s. The error response was: %s.',
          $transaction_data['order_id'],
          $decoded_data['responseMessage'],
        ));
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Merchant Warrior processTokenAuthorization(): Unable to authorize the access token for Order ID: %s. The error response: %message', [
        '%s' => $transaction_data['order_id'],
        '%message' => $e->getMessage(),
      ]);
      throw new PaymentGatewayException($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function processTokenCard(array $transaction_data): array {
    $config = $this->config->get('configuration');
    $merchant_uuid = $config['merchant_uuid'];
    $merchant_api_key = $config['api_key'];
    $merchant_api_passphrase = $config['api_passphrase'];

    // Generate the transaction hash.
    $hash = $this->generateHash([
      md5($merchant_api_passphrase),
      $merchant_uuid,
      $transaction_data['transactionAmount'],
      $transaction_data['transactionCurrency'],
    ]);

    try {
      $options = [
        'form_params' => [
          'method' => 'processCard',
          'merchantUUID' => $merchant_uuid,
          'apiKey' => $merchant_api_key,
          'hash' => urlencode($hash),
        ],
        'headers' => self::HEADERS,
      ];
      // Append the transaction data received to the options form_params.
      $options['form_params'] = $options['form_params'] + $transaction_data;
      $response = $this->httpClient->request('POST', $this->getTokenApiUrl(), $options);
      $data = $response->getBody()->getContents();
      $encode_data = json_encode(simplexml_load_string($data));
      $decoded_data = json_decode($encode_data, TRUE);

      // Return the data if we have a successfuly response.
      if ($decoded_data['responseCode'] === '0') {
        return $decoded_data;
      }
      else {
        $this->logger->error('Merchant Warrior processTokenCard(): Unable to process the card token for Order ID: %s. The error response: %message', [
          '%s' => $transaction_data['order_id'],
          '%message' => $decoded_data['responseMessage'],
        ]);
        throw new HardDeclineException(sprintf('Merchant Warrior: Unable to process the payment for Order ID: %s. The error response was: %s.',
          $transaction_data['order_id'],
          $decoded_data['responseMessage'],
        ));
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Merchant Warrior processTokenCard(): Unable to process the card token for Order ID: %s. The error response: %message', [
        '%s' => $transaction_data['order_id'],
        '%message' => $e->getMessage(),
      ]);
      throw new PaymentGatewayException($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function processCapture(array $transaction_data): array {
    $config = $this->config->get('configuration');
    $merchant_uuid = $config['merchant_uuid'];
    $merchant_api_key = $config['api_key'];
    $merchant_api_passphrase = $config['api_passphrase'];

    // Generate the transaction hash.
    $hash = $this->generateHash([
      md5($merchant_api_passphrase),
      $merchant_uuid,
      $transaction_data['transactionAmount'],
      $transaction_data['transactionCurrency'],
    ]);

    try {
      $options = [
        'form_params' => [
          'method' => 'processCapture',
          'merchantUUID' => $merchant_uuid,
          'apiKey' => $merchant_api_key,
          'hash' => urlencode($hash),
        ],
        'headers' => self::HEADERS,
      ];
      // Append the transaction data received to the options form_params.
      $options['form_params'] = $options['form_params'] + $transaction_data;
      $response = $this->httpClient->request('POST', $this->getApiUrl(), $options);
      $data = $response->getBody()->getContents();
      $encode_data = json_encode(simplexml_load_string($data));
      $decoded_data = json_decode($encode_data, TRUE);

      // Return the data if we have a successfuly response.
      if ($decoded_data['responseCode'] === '0') {
        return $decoded_data;
      }
      else {
        $this->logger->error('Merchant Warrior processCapture(): Unable to capture the transaction for Order ID: %s. The error response: %message', [
          '%s' => $transaction_data['order_id'],
          '%message' => $decoded_data['responseMessage'],
        ]);
        throw new HardDeclineException(sprintf('Merchant Warrior: Unable to capture the payment for Order ID: %s. The error response was: %s.',
          $transaction_data['order_id'],
          $decoded_data['responseMessage'],
        ));
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Merchant Warrior processCapture(): Unable to capture the transaction for Order ID: %s. The error response: %message', [
        '%s' => $transaction_data['order_id'],
        '%message' => $e->getMessage(),
      ]);
      throw new PaymentGatewayException($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function processVoid(array $transaction_data): array {
    $config = $this->config->get('configuration');
    $merchant_uuid = $config['merchant_uuid'];
    $merchant_api_key = $config['api_key'];
    $merchant_api_passphrase = $config['api_passphrase'];

    // Generate the transaction hash.
    $hash = $this->generateHash([
      md5($merchant_api_passphrase),
      $merchant_uuid,
      $transaction_data['transactionID'],
    ]);

    try {
      $options = [
        'form_params' => [
          'method' => 'processVoid',
          'merchantUUID' => $merchant_uuid,
          'apiKey' => $merchant_api_key,
          'hash' => urlencode($hash),
        ],
        'headers' => self::HEADERS,
      ];
      // Append the transaction data received to the options form_params.
      $options['form_params'] = $options['form_params'] + $transaction_data;
      $response = $this->httpClient->request('POST', $this->getApiUrl(), $options);
      $data = $response->getBody()->getContents();
      $encode_data = json_encode(simplexml_load_string($data));
      $decoded_data = json_decode($encode_data, TRUE);

      // Return the data if we have a successfuly response.
      if ($decoded_data['responseCode'] === '0') {
        return $decoded_data;
      }
      else {
        $this->logger->error('Merchant Warrior processVoid(): Unable to process void for Order ID: %s. The error response: %message', [
          '%s' => $transaction_data['order_id'],
          '%message' => $decoded_data['responseMessage'],
        ]);
        throw new HardDeclineException(sprintf('Merchant Warrior: Unable to void the payment for Order ID: %s. The error response was: %s.',
          $transaction_data['order_id'],
          $decoded_data['responseMessage'],
        ));
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Merchant Warrior processVoid(): Unable to process void for Order ID: %s. The error response: %message', [
        '%s' => $transaction_data['order_id'],
        '%message' => $e->getMessage(),
      ]);
      throw new PaymentGatewayException($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function refundCard(array $transaction_data): array {
    $config = $this->config->get('configuration');
    $merchant_uuid = $config['merchant_uuid'];
    $merchant_api_key = $config['api_key'];
    $merchant_api_passphrase = $config['api_passphrase'];

    // Generate the transaction hash.
    $hash = $this->generateHash([
      md5($merchant_api_passphrase),
      $merchant_uuid,
      $transaction_data['transactionAmount'],
      $transaction_data['transactionCurrency'],
    ]);

    try {
      $options = [
        'form_params' => [
          'method' => 'refundCard',
          'merchantUUID' => $merchant_uuid,
          'apiKey' => $merchant_api_key,
          'hash' => urlencode($hash),
        ],
        'headers' => self::HEADERS,
      ];
      // Append the transaction data received to the options form_params.
      $options['form_params'] = $options['form_params'] + $transaction_data;
      $response = $this->httpClient->request('POST', $this->getApiUrl(), $options);
      $data = $response->getBody()->getContents();
      $encode_data = json_encode(simplexml_load_string($data));
      $decoded_data = json_decode($encode_data, TRUE);

      // Return the data if we have a successfuly response.
      if ($decoded_data['responseCode'] === '0') {
        return $decoded_data;
      }
      else {
        $this->logger->error('Merchant Warrior refundCard(): Unable to refund transaction for Order ID: %s. The error response: %message', [
          '%s' => $transaction_data['order_id'],
          '%message' => $decoded_data['responseMessage'],
        ]);
        throw new HardDeclineException(sprintf('Merchant Warrior: Unable to refund the payment for Order ID: %s. The error response was: %s.',
          $transaction_data['order_id'],
          $decoded_data['responseMessage'],
        ));
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Merchant Warrior refundCard(): Unable to refund transaction for Order ID: %s. The error response: %message', [
        '%s' => $transaction_data['order_id'],
        '%message' => $e->getMessage(),
      ]);
      throw new PaymentGatewayException($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function verifyCard(array $transaction_data): array {
    $config = $this->config->get('configuration');
    $merchant_uuid = $config['merchant_uuid'];
    $merchant_api_key = $config['api_key'];

    try {
      $options = [
        'form_params' => [
          'method' => 'verifyCard',
          'merchantUUID' => $merchant_uuid,
          'apiKey' => $merchant_api_key,
        ],
        'headers' => self::HEADERS,
      ];
      // Append the transaction data received to the options form_params.
      $options['form_params'] = $options['form_params'] + $transaction_data;
      // Now, hash the body and append to the header.
      $options['headers']['MW-MESSAGEHASH'] = $this->generateMessageHash($options['form_params'], $config['api_passphrase']);
      $response = $this->httpClient->request('POST', $this->getApiUrl(), $options);
      $data = $response->getBody()->getContents();
      $encode_data = json_encode(simplexml_load_string($data));
      $decoded_data = json_decode($encode_data, TRUE);

      // Return the data if we have a successful response.
      if ($decoded_data['responseCode'] === '0') {
        return $decoded_data;
      }
      else {
        $this->logger->error('Merchant Warrior verifyCard(): Unable to verify a card for Order ID: %s. The error response: %message', [
          '%s' => $transaction_data['order_id'],
          '%message' => $decoded_data['responseMessage'],
        ]);
        throw new HardDeclineException(sprintf('Merchant Warrior: Unable to verify the payment method. The error response was: %s.',
          $decoded_data['responseMessage'],
        ));
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Merchant Warrior verifyCard(): Unable to verify a card for Order ID: %s. The error response: %message', [
        '%s' => $transaction_data['order_id'],
        '%message' => $e->getMessage(),
      ]);
      throw new PaymentGatewayException($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function cardInfo(array $transaction_data): array {
    $config = $this->config->get('configuration');
    $merchant_uuid = $config['merchant_uuid'];
    $merchant_api_key = $config['api_key'];

    try {
      $options = [
        'form_params' => [
          'method' => 'cardInfo',
          'merchantUUID' => $merchant_uuid,
          'apiKey' => $merchant_api_key,
        ],
        'headers' => self::HEADERS,
      ];
      // Append the transaction data received to the options form_params.
      $options['form_params'] = $options['form_params'] + $transaction_data;
      $response = $this->httpClient->request('POST', $this->getTokenApiUrl(), $options);
      $data = $response->getBody()->getContents();
      $encode_data = json_encode(simplexml_load_string($data));
      $decoded_data = json_decode($encode_data, TRUE);

      // Return the data if we have a successful response.
      if ($decoded_data['responseCode'] === '0') {
        return $decoded_data;
      }
      else {
        $this->logger->error('Merchant Warrior cardInfo(): Unable to get card info for Order ID: %s. The error response: %message', [
          '%s' => $transaction_data['order_id'],
          '%message' => $decoded_data['responseMessage'],
        ]);
        throw new HardDeclineException(sprintf('Merchant Warrior: Unable to fetch additional card details for the payment method. The error response was: %s.',
          $decoded_data['responseMessage'],
        ));
      }
    }
    catch (\Exception $e) {
      $this->logger->error('Merchant Warrior cardInfo(): Unable to get card info for Order ID: %s. The error response: %message', [
        '%s' => $transaction_data['order_id'],
        '%message' => $e->getMessage(),
      ]);
      throw new PaymentGatewayException($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getApiUrl(): string {
    $config = $this->config->get('configuration');
    return $config['mode'] === 'test' ? self::MW_DIRECT_API_TEST_URL : self::MW_DIRECT_API_PROD_URL;
  }

  /**
   * {@inheritdoc}
   */
  public function getTokenApiUrl(): string {
    $config = $this->config->get('configuration');
    return $config['mode'] === 'test' ? self::MW_TOKEN_API_TEST_URL : self::MW_TOKEN_API_PROD_URL;
  }

  /**
   * {@inheritdoc}
   */
  public function getPayframeSubmitUrl(): string {
    $config = $this->config->get('configuration');
    return $config['mode'] === 'test' ? self::MW_PAYFRAME_SUBMIT_TEST_URL : self::MW_PAYFRAME_SUBMIT_PROD_URL;
  }

  /**
   * {@inheritdoc}
   */
  public function getPayframeSrcUrl(): string {
    $config = $this->config->get('configuration');
    return $config['mode'] === 'test' ? self::MW_PAYFRAME_SRC_TEST_URL : self::MW_PAYFRAME_SRC_PROD_URL;
  }

  /**
   * Generate the hash.
   *
   * @param array $params
   *   The parameters to hash. ie. API passphrase, UUID, amount, currency.
   *
   * @return string
   *   The generated hash.
   */
  private function generateHash(array $params): string {
    $param_string = '';
    foreach ($params as $param) {
      $param_string .= $param;
    }
    $parts = strtolower($param_string);
    return md5($parts);
  }

  /**
   * Generate the message header hash.
   *
   * @param array $params
   *   The parameters to hash. ie. API passphrase, UUID, amount, currency.
   * @param string $api_passphrase
   *   The API passphrase.
   *
   * @return string
   *   The generated hash.
   */
  private function generateMessageHash(array $params, string $api_passphrase): string {
    // Sort the parameters alphabetically by key.
    ksort($params);

    // Build a query string from the sorted parameters.
    $query_string = http_build_query($params, '', '&', PHP_QUERY_RFC1738);
    $query_string = urldecode($query_string);

    // Hash the query string with the API passphrase using HMAC-SHA256.
    $hash = hash_hmac('sha256', $query_string, $api_passphrase);

    return strtolower($hash);
  }

}
