<?php

namespace Drupal\commerce_merchant_warrior\Api;

/**
 * An interface for the MerchantWarriorApi service.
 */
interface MerchantWarriorApiInterface {

  /**
   * Merchant Warrior Direct test API URL.
   */
  const MW_DIRECT_API_TEST_URL = 'https://base.merchantwarrior.com/post/';

  /**
   * Merchant Warrior Direct production API URL.
   */
  const MW_DIRECT_API_PROD_URL = 'https://api.merchantwarrior.com/post/';

  /**
  * Merchant Warrior Token test API URL.
  */
  const MW_TOKEN_API_TEST_URL = 'https://base.merchantwarrior.com/token/';

  /**
   * Merchant Warrior Token production API URL.
   */
  const MW_TOKEN_API_PROD_URL = 'https://api.merchantwarrior.com/token/';

  /**
   * Merchant Warrior Payframe test submit URL.
   */
  const MW_PAYFRAME_SUBMIT_TEST_URL = 'https://base.merchantwarrior.com/payframe/';

  /**
   * Merchant Warrior Payframe production submit URL.
   */
  const MW_PAYFRAME_SUBMIT_PROD_URL = 'https://api.merchantwarrior.com/payframe/';

  /**
   * Merchant Warrior Payframe test src URL.
   */
  const MW_PAYFRAME_SRC_TEST_URL = 'https://securetest.merchantwarrior.com/payframe/';

  /**
   * Merchant Warrior Payframe test src URL.
   */
  const MW_PAYFRAME_SRC_PROD_URL = 'https://secure.merchantwarrior.com/payframe/';

  /**
   * Creates and returns an access token from Merchant Warrior's Direct API.
   *
   * @return string
   *   The API access token.
   */
  public function getAccessToken(): string;

  /**
   * Authorizes the payment using the Merchant Warrior Direct API.
   *
   * @param array $transaction_data
   *   An array of transaction data for the card to pass to the MW API.
   *
   * @return array
   *   The response array.
   */
  public function processAuthorization(array $transaction_data) : array;

  /**
   * Authorizes the payment using the Merchant Warrior Token API.
   *
   * @param array $transaction_data
   *   An array of transaction data for the card to pass to the MW API.
   *
   * @return array
   *   The response array.
   */
  public function processTokenAuthorization(array $transaction_data) : array;

  /**
   * Process a complete payment using the Merchant Warrior Token API.
   *
   * It authorizes and captures the funds immediately.
   *
   * This would be identical to saving a customer.
   *
   * @param array $transaction_data
   *   An array of transaction data for the card to pass to the MW API.
   *
   * @return array
   *   The response array.
   */
  public function processTokenCard(array $transaction_data) : array;

  /**
   * Capture pre-authorized funds using the Merchant Warrior Direct API.
   *
   * @param array $transaction_data
   *   An array of transaction data for the card to pass to the MW API.
   *
   * @return array
   *   The response array.
   */
  public function processCapture(array $transaction_data) : array;

  /**
   * Perform a void on a transaction authorized by Merchant Warrior Direct API.
   *
   * @param array $transaction_data
   *   An array of transaction data for the card to pass to the MW API.
   *
   * @return array
   *   The response array.
   */
  public function processVoid(array $transaction_data) : array;

  /**
   * Verify a payment method using the Merchant Warrior Direct API.
   *
   * @param array $transaction_data
   *   An array of transaction data for the card to pass to the MW API.
   *
   * @return array
   *   The response array.
   */
  public function verifyCard(array $transaction_data) : array;

  /**
   * Queries an existing card for info using the Merchant Warrior Direct API.
   *
   * @param array $transaction_data
   *   An array of transaction data about the card to pass to the MW API.
   *
   * @return array
   *   The response array.
   */
  public function cardInfo(array $transaction_data) : array;

  /**
   * Refund a payment captured by the Merchant Warrior Direct API.
   *
   * @param array $transaction_data
   *   An array of transaction data for the card to pass to the MW API.
   *
   * @return array
   *   The response array.
   */
  public function refundCard(array $transaction_data) : array;

  /**
   * Returns the Merchant Warrior Direct API URL.
   *
   * @return string
   *   The API url.
   */
  public function getApiUrl(): string;

  /**
   * Returns the Merchant Warrior Token API URL.
   *
   * @return string
   *   The API url.
   */
  public function getTokenApiUrl(): string;

  /**
   * Returns the Merchant Warrior Payframe submit URL.
   *
   * @return string
   *   The submit url.
   */
  public function getPayframeSubmitUrl(): string;

  /**
   * Returns the Merchant Warrior Payframe submit URL.
   *
   * @return string
   *   The src url.
   */
  public function getPayframeSrcUrl(): string;

}
