<?php

namespace Drupal\commerce_merchant_warrior\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_price\Price;
use Drupal\commerce_payment\CreditCard;
use Drupal\Core\Form\FormStateInterface;
use GuzzleHttp\Exception\RequestException;
use Drupal\commerce_price\RounderInterface;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_price\MinorUnitsConverterInterface;
use Drupal\commerce_payment\Entity\PaymentMethodInterface;
use Drupal\commerce_payment\Exception\HardDeclineException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\commerce_payment\Exception\InvalidRequestException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OnsitePaymentGatewayBase;

/**
 * Provides the Merchant Warrior Payframe payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "merchant_warrior_payframe",
 *   label = "Merchant Warrior Payframe",
 *   display_label = "Merchant Warrior Payframe",
 *   forms = {
 *     "add-payment-method" = "Drupal\commerce_merchant_warrior\PluginForm\Onsite\PayframePaymentForm",
 *   },
 *   payment_method_types = {"credit_card"},
 *   credit_card_types = {
 *     "amex", "dinersclub", "discover", "mastercard", "visa",
 *   },
 *   requires_billing_information = TRUE,
 * )
 */
class Payframe extends OnsitePaymentGatewayBase implements PayframeInterface {

  /**
   * The Merchant Warrior API client.
   *
   * @var \Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface
   */
  protected MerchantWarriorApiInterface $apiClient;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The rounder.
   *
   * @var \Drupal\commerce_price\RounderInterface
   */
  protected RounderInterface $rounder;

  /**
   * The current route.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected RouteMatchInterface $currentRoute;

  /**
   * Constructs a new Payframe payment gateway object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   The payment type manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   The payment method type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   * @param \Drupal\commerce_price\MinorUnitsConverterInterface $minor_units_converter
   *   The minor units converter.
   * @param \Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface $api_client
   *   The Merchant Warrior API client.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger.
   * @param \Drupal\commerce_price\RounderInterface $rounder
   *   The rounder.
   * @param \Drupal\Core\Routing\RouteMatchInterface $current_route
   *   The current route.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    PaymentTypeManager $payment_type_manager,
    PaymentMethodTypeManager $payment_method_type_manager,
    TimeInterface $time,
    ?MinorUnitsConverterInterface $minor_units_converter = NULL,
    MerchantWarriorApiInterface $api_client,
    MessengerInterface $messenger,
    RounderInterface $rounder,
    RouteMatchInterface $current_route
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time, $minor_units_converter);

    $this->apiClient = $api_client;
    $this->messenger = $messenger;
    $this->rounder = $rounder;
    $this->currentRoute = $current_route;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('commerce_price.minor_units_converter'),
      $container->get('commerce_merchant_warrior.api_client'),
      $container->get('messenger'),
      $container->get('commerce_price.rounder'),
      $container->get('current_route_match')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'merchant_uuid' => '',
      'api_key' => '',
      'api_passphrase' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['merchant_uuid'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Merchant UUID'),
      '#default_value' => $this->configuration['merchant_uuid'],
      '#required' => TRUE,
    ];
    $form['api_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('API Key'),
      '#default_value' => $this->configuration['api_key'],
      '#required' => TRUE,
    ];
    $form['api_passphrase'] = [
      '#type' => 'textfield',
      '#title' => $this->t('API Passphrase'),
      '#default_value' => $this->configuration['api_passphrase'],
      '#required' => TRUE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    $values = $form_state->getValue($form['#parents']);
    $this->configuration['merchant_uuid'] = $values['merchant_uuid'];
    $this->configuration['api_key'] = $values['api_key'];
    $this->configuration['api_passphrase'] = $values['api_passphrase'];
  }

  /**
   * {@inheritdoc}
   */
  public function capturePayment(PaymentInterface $payment, ?Price $amount = NULL) {
    $this->assertPaymentState($payment, ['authorization']);

    // If not specified, capture the entire amount.
    $amount = $amount ?: $payment->getAmount();
    $amount = $this->rounder->round($amount);

    try {
      $transaction_data = $this->apiClient->processCapture([
        'transactionAmount' => $amount->getNumber(),
        'transactionCurrency' => $amount->getCurrencyCode(),
        'captureAmount' => $amount->getNumber(),
        'transactionID' => $payment->getRemoteId(),
        'order_id' => $payment->getOrder()->id(),
      ]);

      $payment->setState('completed');
      $payment->setRemoteId($transaction_data['transactionID']);
      $payment->save();
    }
    catch (RequestException $e) {
      throw new PaymentGatewayException('Could not capture payment. Message: ' . $e->getMessage(), $e->getCode(), $e);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createPayment(PaymentInterface $payment, $capture = TRUE) {
    $this->assertPaymentState($payment, ['new']);
    $payment_method = $payment->getPaymentMethod();
    $this->assertPaymentMethod($payment_method);

    // Get the ammount.
    $amount = $payment->getAmount();
    $amount = $this->rounder->round($amount);

    // Get the order and address.
    $order = $payment->getOrder();
    /** @var \Drupal\address\Plugin\Field\FieldType\AddressItem $address */
    $address = $order->getBillingProfile()->get('address')->first();

    try {
      $data = [
        'transactionAmount' => $amount->getNumber(),
        'transactionCurrency' => $amount->getCurrencyCode(),
        'transactionProduct' => 'Order ID: ' . $order->id(),
        'customerEmail' => $order->getEmail(),
        'customerName' => $address->getGivenName() . ' ' . $address->getFamilyName(),
        'customerAddress' => $address->getAddressLine1(),
        'customerCity' => $address->getLocality(),
        'customerState' => $address->getAdministrativeArea(),
        'customerPostCode' => $address->getPostalCode(),
        'customerCountry' => $address->getCountryCode(),
        'cardID' => $payment_method->getRemoteId(),
        'order_id' => $order->id(),
      ];
      if ($capture) {
        $data += ['captureAmount' => $amount->getNumber()];
        $transaction_data = $this->apiClient->processTokenCard($data);
      }
      else {
        $transaction_data = $this->apiClient->processTokenAuthorization($data);
      }

      // Mark the payment as completed.
      $next_state = $capture ? 'completed' : 'authorization';
      if (!$capture) {
        $payment->setExpiresTime($this->time->getRequestTime() + (86400 * 29));
      }
      $payment->setState($next_state);
      $payment->setRemoteId($transaction_data['transactionID']);
      $payment->save();
    }
    catch (RequestException $e) {
      throw new HardDeclineException('Could not charge the payment method.');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createPaymentMethod(PaymentMethodInterface $payment_method, array $payment_details) {
    $required_keys = [
      'merchant_warrior_access_token',
      'merchant_warrior_payframe_token',
      'merchant_warrior_payframe_key',
      'currency_code',
    ];

    foreach ($required_keys as $required_key) {
      if (empty($payment_details[$required_key])) {
        throw new \InvalidArgumentException(
          sprintf('$payment_details must contain the %s key.', $required_key)
        );
      }
    }

    try {
      // Get the cart from the current route.
      /** @var \Drupal\commerce_order\Entity\OrderInterface $order */
      $order = $this->currentRoute->getParameter('commerce_order');
      /** @var \Drupal\address\Plugin\Field\FieldType\AddressItem $address */
      $address = $payment_method->getBillingProfile()->get('address')->first();

      $transaction_data = $this->apiClient->verifyCard([
        'transactionCurrency' => $order->getTotalPrice()->getCurrencyCode(),
        'customerName' => $address->getGivenName() . ' ' . $address->getFamilyName(),
        'customerAddress' => $address->getAddressLine1(),
        'customerCity' => $address->getLocality(),
        'customerState' => $address->getAdministrativeArea(),
        'customerPostCode' => $address->getPostalCode(),
        'customerCountry' => $address->getCountryCode(),
        'accessToken' => $payment_details['merchant_warrior_access_token'],
        'payframeToken' => $payment_details['merchant_warrior_payframe_token'],
        'payframeKey' => $payment_details['merchant_warrior_payframe_key'],
        'addCard' => 1,
        'order_id' => $order->id(),
      ]);

      // Make another call to fetch the required additional card details.
      $card_data = $this->apiClient->cardInfo([
        'order_id' => $order->id(),
        'cardID' => $transaction_data['cardID'],
      ]);

      // Save the card details to the payment method.
      $card_number = $card_data['cardBin'] . '000000' . $card_data['cardNumberLast'];
      // Extract the card type.
      $type = CreditCard::detectType($card_number);
      $payment_method->card_type = $type->getId();
      // Only the last 4 numbers are safe to store.
      $payment_method->card_number = $card_data['cardNumberLast'];
      $payment_method->card_exp_month = $card_data['cardExpiryMonth'];
      $payment_method->card_exp_year = $card_data['cardExpiryYear'];

      // Calculate the expiration timestamp.
      $expires = CreditCard::calculateExpirationTimestamp(
        $card_data['cardExpiryMonth'],
        $card_data['cardExpiryYear']
      );
      // Store the remote ID returned by the request.
      $payment_method
        ->setRemoteId($card_data['cardID'])
        ->setExpiresTime($expires)
        ->save();
    }
    catch (RequestException $e) {
      throw new HardDeclineException("Unable to store the credit card.");
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deletePaymentMethod(PaymentMethodInterface $payment_method) {
    // Delete the remote record here, throw an exception if it fails.
    // See \Drupal\commerce_payment\Exception for the available exceptions.
    // Delete the local entity.
    $payment_method->delete();
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(PaymentInterface $payment, ?Price $amount = NULL) {
    $this->assertPaymentState($payment, ['completed', 'partially_refunded']);
    if ($payment->getCompletedTime() < strtotime('-180 days')) {
      throw new InvalidRequestException("Unable to refund a payment captured more than 180 days ago.");
    }
    // If not specified, refund the entire amount.
    $amount = $amount ?: $payment->getAmount();
    $amount = $this->rounder->round($amount);
    $this->assertRefundAmount($payment, $amount);
    $remote_id = $payment->getRemoteId();
    if (empty($remote_id)) {
      throw new InvalidRequestException('Could not determine the remote payment details.');
    }

    try {
      $new_refunded_amount = $payment->getRefundedAmount()->add($amount);

      $this->apiClient->refundCard([
        'transactionAmount' => $amount->getNumber(),
        'transactionCurrency' => $amount->getCurrencyCode(),
        'refundAmount' => $amount->getNumber(),
        'transactionID' => $remote_id,
        'order_id' => $payment->getOrder()->id(),
      ]);

      if ($new_refunded_amount->lessThan($payment->getAmount())) {
        $payment->setState('partially_refunded');
      }
      else {
        $payment->setState('refunded');
      }

      $payment->setRefundedAmount($new_refunded_amount);
      $payment->save();
    }
    catch (RequestException $e) {
      throw new InvalidRequestException("Could not refund the payment.", $e->getCode(), $e);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function voidPayment(PaymentInterface $payment) {
    $this->assertPaymentState($payment, ['authorization']);

    $remote_id = $payment->getRemoteId();
    if (empty($remote_id)) {
      throw new PaymentGatewayException('Remote authorization ID could not be determined.');
    }

    // Void the transaction on Merchant Warrior.
    try {
      $this->apiClient->processVoid([
        'transactionID' => $remote_id,
        'order_id' => $payment->getOrder()->id(),
      ]);

      $payment->setState('authorization_voided');
      $payment->save();
    }
    catch (RequestException $e) {
      throw new \InvalidArgumentException('Only payments in the "authorization" state can be voided.');
    }
  }

}
