<?php

namespace Drupal\commerce_merchant_warrior\Plugin\rest\resource;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\CreditCard;
use Drupal\commerce_price\RounderInterface;
use Drupal\rest\Plugin\ResourceBase;
use Drupal\rest\ResourceResponse;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a way to process an authorization with Merchant Warrior.
 *
 * @RestResource(
 *   id = "merchant_warrior_process_authorization",
 *   label = @Translation("Merchant Warrior: Process an authorization request resource"),
 *   uri_paths = {
 *     "create" = "/api/merchant-warrior/process-authorization"
 *   }
 * )
 */
class MerchantWarriorProcessAuthorization extends ResourceBase {

  use StringTranslationTrait;

  /**
   * The Merchant Warrior API client.
   *
   * @var \Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface
   */
  protected MerchantWarriorApiInterface $apiClient;

  /**
   * The rounder.
   *
   * @var \Drupal\commerce_price\RounderInterface
   */
  protected RounderInterface $rounder;

  /**
   * The time.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected TimeInterface $time;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Constructs a MerchantWarriorProcessAuthorization instance.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param array $serializer_formats
   *   The available serialization formats.
   * @param \Psr\Log\LoggerInterface $logger
   *   A logger instance.
   * @param \Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface $api_client
   *   The Merchant Warrior API client.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_price\RounderInterface $rounder
   *   The rounder.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    $serializer_formats,
    LoggerInterface $logger,
    MerchantWarriorApiInterface $api_client,
    EntityTypeManagerInterface $entity_type_manager,
    RounderInterface $rounder,
    TimeInterface $time,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $serializer_formats, $logger);

    $this->apiClient = $api_client;
    $this->entityTypeManager = $entity_type_manager;
    $this->rounder = $rounder;
    $this->time = $time;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->getParameter('serializer.formats'),
      $container->get('logger.factory')->get('rest'),
      $container->get('commerce_merchant_warrior.api_client'),
      $container->get('entity_type.manager'),
      $container->get('commerce_price.rounder'),
      $container->get('datetime.time'),
    );
  }

  /**
   * Responds to POST request to process an authorization with Merchant Warrior.
   *
   * Process an authorization on the requested amount using the token and
   * encryption key.
   *
   * @return \Drupal\rest\ResourceResponse
   *   The authorization data.
   */
  public function post(array $data = []): ResourceResponse {
    if (empty($data['merchant_warrior_access_token'])
      || empty($data['merchant_warrior_payframe_token'])
      || empty($data['merchant_warrior_payframe_key'])
      || empty($data['order_uuid'])
    ) {
      return new ResourceResponse($this->t('Merchant Warrior: required keys are missing from the body parameters.'), 401);
    }

    try {
      /** @var \Drupal\commerce_order\Entity\OrderInterface[] $orders */
      $orders = $this->entityTypeManager
        ->getStorage('commerce_order')
        ->loadByProperties(['uuid' => $data['order_uuid']]);

      if (empty($orders)) {
        $this->logger->error(sprintf('Merchant Warrior: The order %s does not exist.', $data['order_uuid']));
        return new ResourceResponse($this->t('Merchant Warrior: The order @order_id does not exist.', ['@order_id' => $data['order_uuid']]), 500);
      }

      $order = current($orders);
      /** @var \Drupal\address\Plugin\Field\FieldType\AddressItem $address */
      $address = $order->getBillingProfile()->get('address')->first();

      // Process the authorization.
      $amount = $order->getTotalPrice();
      $amount = $this->rounder->round($amount);
      $transaction_data = $this->apiClient->processAuthorization([
        'transactionAmount' => $amount->getNumber(),
        'transactionCurrency' => $order->getTotalPrice()->getCurrencyCode(),
        'transactionProduct' => 'Order ID: ' . $order->id(),
        'customerEmail' => $order->getEmail(),
        'customerName' => $address->getGivenName() . ' ' . $address->getFamilyName(),
        'customerAddress' => $address->getAddressLine1(),
        'customerCity' => $address->getLocality(),
        'customerState' => $address->getAdministrativeArea(),
        'customerPostCode' => $address->getPostalCode(),
        'customerCountry' => $address->getCountryCode(),
        'addCard' => 1,
        'accessToken' => $data['merchant_warrior_access_token'],
        'payframeToken' => $data['merchant_warrior_payframe_token'],
        'payframeKey' => $data['merchant_warrior_payframe_key'],
        'order_id' => $order->id(),
      ]);

      // Create a payment object and attach it to the order and return back the
      // order details.
      if (!empty($transaction_data['transactionID']) && !empty($transaction_data['cardID'])) {
        $created = $this->createOrderPayment($transaction_data, $order);
        if ($created) {
          $return_data = [
            'order_id' => $order->id(),
            'order_uuid' => $order->uuid(),
            'order_number' => $order->getOrderNumber(),
            // Status of 0 indicates to the FE that this was a successful
            // payment.
            'status' => 0,
          ];

          return new ResourceResponse($return_data);
        }
      }
      return new ResourceResponse($this->t('Merchant Warrior: Payment authorization failed for order.'), 500);
    }
    // If there were errors while creating the order payment.
    catch (\Exception $e) {
      $message = 'Merchant Warrior: Payment authorization failed. The response received from the API is: <pre>@response</pre>';
      $this->logger->error($e->getMessage());

      return new ResourceResponse([
        'status' => -1,
        'message' => $message,
      ], 500);
    }
  }

  /**
   * Create a payment method and payment object and attach it to the order.
   *
   * @param array $transaction_data
   *   The transaction data from the API.
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order object.
   *
   * @return bool
   *   The status of the payment creation.
   */
  public function createOrderPayment(array $transaction_data, OrderInterface $order): bool {
    $uid = $order->getCustomerId();

    // Load the Payframe payment gateway plugin.
    /** @var \Drupal\commerce_payment\Entity\PaymentGatewayInterface $payment_gateway */
    $payment_gateway = $this->entityTypeManager->getStorage('commerce_payment_gateway')->load('merchant_warrior_payframe');

    // Get the card expiration time.
    $expire_year = \DateTime::createFromFormat('y', (string) $transaction_data['cardExpiryYear']);
    $expire_year = $expire_year->format('Y');
    $expires = CreditCard::calculateExpirationTimestamp(
        (string) $transaction_data['cardExpiryMonth'],
        $expire_year
      );
    $card_number = $transaction_data['paymentCardNumber'];
    // Convert the masked card to a numeric one. The card number will be
    // like this: 555550XXXXXX5439, so, the first 6 digits and last 4.
    // We just need to make sure the card number is not a string so, we need
    // to convert "X" into a number for ex: 0. This way we'll be able to get
    // the card type.
    $card_number = str_replace('X', '0', $card_number);
    // Extract the card type.
    $type = CreditCard::detectType($card_number);

    /** @var \Drupal\commerce_payment\Entity\PaymentMethodInterface $payment_method */
    $payment_method = $this->entityTypeManager->getStorage('commerce_payment_method')->create([
      'uid' => $uid,
      'type' => 'credit_card',
      'payment_gateway' => $payment_gateway->id(),
      'card_type' => $type->getId(),
      'card_number' => substr($card_number, -4),
      'billing_profile' => $order->getBillingProfile(),
      'card_exp_month' => $transaction_data['cardExpiryMonth'],
      'card_exp_year' => $expire_year,
      'reusable' => TRUE,
    ]);
    $payment_method->setExpiresTime($expires);
    $payment_method->setRemoteId($transaction_data['cardID']);
    $payment_method->save();

    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $this->entityTypeManager->getStorage('commerce_payment')->create([
      'state' => 'authorization',
      'payment_method' => $payment_method->id(),
      'amount' => $order->getTotalPrice(),
      'payment_gateway' => $payment_gateway->id(),
      'order_id' => $order->id(),
      'remote_id' => $transaction_data['transactionID'],
      'payment_gateway_mode' => $payment_gateway->getPlugin()->getMode(),
      'uid' => $uid,
    ]);
    $payment->setExpiresTime($this->time->getRequestTime() + (86400 * 29));
    $payment->save();

    return TRUE;
  }

}
