<?php

namespace Drupal\commerce_merchant_warrior\PluginForm\Onsite;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\commerce\InlineFormManager;
use Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface;
use Drupal\commerce_payment\PluginForm\PaymentMethodAddForm;
use Drupal\commerce_store\CurrentStoreInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides onsite payment form via Merchant Warrior's Payframe integration.
 */
class PayframePaymentForm extends PaymentMethodAddForm {

  /**
   * The Merchant Warrior API client.
   *
   * @var \Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface
   */
  protected $apiClient;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Constructs a new PayframePaymentForm object.
   *
   * @param \Drupal\commerce_store\CurrentStoreInterface $current_store
   *   The current store.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce\InlineFormManager $inline_form_manager
   *   The inline form manager.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   * @param \Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface $api_client
   *   The Merchant Warrior API client.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger.
   */
  public function __construct(CurrentStoreInterface $current_store, EntityTypeManagerInterface $entity_type_manager, InlineFormManager $inline_form_manager, LoggerInterface $logger, MerchantWarriorApiInterface $api_client, MessengerInterface $messenger) {
    parent::__construct(
      $current_store,
      $entity_type_manager,
      $inline_form_manager,
      $logger
    );

    $this->apiClient = $api_client;
    $this->messenger = $messenger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('commerce_store.current_store'),
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_inline_form'),
      $container->get('logger.channel.commerce_merchant_warrior'),
      $container->get('commerce_merchant_warrior.api_client'),
      $container->get('messenger'),
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function buildCreditCardForm(array $element, FormStateInterface $form_state) {
    // To display the card data.
    $element['mw_card_data'] = [
      '#type' => 'markup',
      '#markup' => '<div id="cardData"></div>',
      '#weight' => -201,
    ];

    // To display validation errors.
    $element['payment_errors'] = [
      '#type' => 'markup',
      '#markup' => '<div id="payment-errors"></div>',
      '#weight' => -200,
    ];

    // Generate an access token from Merchant Warrior to initiate the Payframe
    // session.
    if (empty($form_state->getValue('merchant_warrior_access_token'))) {
      $plugin = $this->entity->getPaymentGateway()->getPlugin();
      $config = $plugin->getConfiguration();
      $access_token = $this->apiClient->getAccessToken();
    }
    else {
      $access_token = $form_state->getValue('merchant_warrior_access_token');
    }
    $element['merchant_warrior_access_token'] = [
      '#type' => 'hidden',
      '#attributes' => [
        'id' => 'merchant-warrior-access-token',
      ],
    ];
    $element['merchant_warrior_payframe_token'] = [
      '#type' => 'hidden',
      '#attributes' => [
        'id' => 'merchant-warrior-payframe-token',
      ],
    ];
    $element['merchant_warrior_payframe_key'] = [
      '#type' => 'hidden',
      '#attributes' => [
        'id' => 'merchant-warrior-payframe-key',
      ],
    ];
    $element['currency_code'] = [
      '#type' => 'value',
      '#value' => $this->currentStore->getStore()->getDefaultCurrencyCode(),
    ];

    // Include the necessary libraries and settings.
    $submit_url = $config['mode'] === 'test' ? MerchantWarriorApiInterface::MW_PAYFRAME_SUBMIT_TEST_URL : MerchantWarriorApiInterface::MW_PAYFRAME_SUBMIT_PROD_URL;
    $src = $config['mode'] === 'test' ? MerchantWarriorApiInterface::MW_PAYFRAME_SRC_TEST_URL : MerchantWarriorApiInterface::MW_PAYFRAME_SRC_PROD_URL;
    $element['#attributes']['class'][] = 'commerce-merchant-warrior-payframe-form';
    $element['#attached']['drupalSettings']['commerceMerchantWarriorPayframe'] = [
      'submitUrl' => $submit_url,
      'accessToken' => $access_token,
      'src' => $src,
    ];

    // Attach the Merchant Warrior library depending on the mode.
    $element['#attached']['library'][] = 'commerce_merchant_warrior/merchant_warrior_payframe_test';
    if ($config['mode'] === 'test') {
      $element['#attached']['library'][] = 'commerce_merchant_warrior/merchant_warrior_payframe_prod';
    }
    // Attach the Payframe form library.
    $element['#attached']['library'][] = 'commerce_merchant_warrior/payframe_form';

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  protected function validateCreditCardForm(
    array &$element,
    FormStateInterface $form_state,
  ) {
    // The JS library performs its own validation.
  }

  /**
   * {@inheritdoc}
   */
  protected function submitCreditCardForm(
    array $element,
    FormStateInterface $form_state,
  ) {
    // The payment gateway plugin will process the submitted payment details.
  }

}
