<?php

namespace Drupal\Tests\commerce_merchant_warrior\Unit\Plugin\rest\resource;

use CommerceGuys\Addressing\Address;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Tests\UnitTestCase;
use Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface;
use Drupal\commerce_merchant_warrior\Plugin\rest\resource\MerchantWarriorProcessAuthorization;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentGatewayInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Entity\PaymentMethodInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\PaymentGatewayInterface as PaymentGatewayPaymentGatewayInterface;
use Drupal\commerce_price\Price;
use Drupal\commerce_price\RounderInterface;
use Drupal\profile\Entity\ProfileInterface;
use Drupal\rest\ResourceResponse;
use Psr\Log\LoggerInterface;

/**
 * @coversDefaultClass \Drupal\commerce_merchant_warrior\Plugin\rest\resource\MerchantWarriorProcessAuthorization
 */
class MerchantWarriorProcessAuthorizationTest extends UnitTestCase {

  use StringTranslationTrait;

  /**
   * The API client mock.
   *
   * @var \PHPUnit\Framework\MockObject\MockObject|\Drupal\commerce_merchant_warrior\Api\MerchantWarriorApiInterface
   */
  protected $apiClient;

  /**
   * The entity type manager mock.
   *
   * @var \PHPUnit\Framework\MockObject\MockObject|\Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The logger mock.
   *
   * @var \PHPUnit\Framework\MockObject\MockObject|\Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The resource instance.
   *
   * @var \Drupal\commerce_merchant_warrior\Plugin\rest\resource\MerchantWarriorProcessAuthorization
   */
  protected $resource;

  /**
   * The container.
   *
   * @var \Drupal\Core\DependencyInjection\ContainerBuilder
   */
  protected $container;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    // Create mocks for the dependencies.
    $this->apiClient = $this->createMock(MerchantWarriorApiInterface::class);
    $this->entityTypeManager = $this->createMock(EntityTypeManagerInterface::class);
    $this->logger = $this->createMock(LoggerInterface::class);
    $this->rounder = $this->createMock(RounderInterface::class);
    $this->rounder->method('round')->willReturnCallback(function (Price $price) {
      return new Price(round($price->getNumber(), 2), $price->getCurrencyCode());
    });
    $this->time = $this->createMock(TimeInterface::class);

    // Set up the service mocks in the container.
    $this->container = new ContainerBuilder();
    $this->container->set('commerce_merchant_warrior.api_client', $this->apiClient);
    $this->container->set('entity_type.manager', $this->entityTypeManager);
    $this->container->set('logger.channel.default', $this->logger);
    $this->container->set('string_translation', $this->createMock(TranslationInterface::class));
    \Drupal::setContainer($this->container);

    // Initialize the resource with mocked dependencies.
    $configuration = [];
    $plugin_id = 'merchant_warrior_process_authorization';
    $plugin_definition = [];
    $serializer_formats = ['json'];

    $this->resource = new MerchantWarriorProcessAuthorization(
        $configuration, $plugin_id, $plugin_definition, $serializer_formats,
        $this->logger, $this->apiClient, $this->entityTypeManager,
        $this->rounder, $this->time
    );
  }

  /**
   * Tests the POST method with valid data.
   */
  public function testPostSuccess(): void {
    // Set up the mocks for the entity storage.
    $this->setMocks();

    // Mock the processAuthorization method to return the expected data.
    $this->apiClient->expects($this->once())
      ->method('processAuthorization')
      ->willReturn([
        'transactionID' => 'transaction_id',
        'cardID' => 'card_id',
        'cardExpiryMonth' => '12',
        'cardExpiryYear' => '24',
        'paymentCardNumber' => '555550XXXXXX5439',
      ]);

    $response = $this->resource->post([
      'merchant_warrior_access_token' => 'access_token',
      'merchant_warrior_payframe_token' => 'payframe_token',
      'merchant_warrior_payframe_key' => 'payframe_key',
      'order_uuid' => '1234',
    ]);

    $this->assertInstanceOf(ResourceResponse::class, $response);
    $this->assertEquals(200, $response->getStatusCode());
    $this->assertEquals([
      'order_id' => 1,
      'order_uuid' => '1234',
      'order_number' => NULL,
      'status' => 0,
    ], $response->getResponseData());
  }

  /**
   * Tests the POST method with missing parameters.
   */
  public function testPostMissingParameters(): void {
    $response = $this->resource->post([]);

    $this->assertInstanceOf(ResourceResponse::class, $response);
    $this->assertEquals(401, $response->getStatusCode());
    $this->assertEquals((string) $this->t('Merchant Warrior: required keys are missing from the body parameters.'), $response->getResponseData());
  }

  /**
   * Tests the POST method when the order is not found.
   */
  public function testPostOrderNotFound(): void {
    // Mock the storage of 'commerce_order' to return an empty array.
    $order_storage = $this->createMock(EntityStorageInterface::class);
    $order_storage->expects($this->any())
      ->method('loadByProperties')
      ->with(['uuid' => 'non_existent_order_uuid'])
      ->willReturn(NULL);

    // Set up the mocks for the entity storage.
    $this->setMocks($order_storage);

    $this->logger->expects($this->once())
      ->method('error')
      ->with($this->stringContains('The order'));

    $response = $this->resource->post([
      'merchant_warrior_access_token' => 'access_token',
      'merchant_warrior_payframe_token' => 'payframe_token',
      'merchant_warrior_payframe_key' => 'payframe_key',
      'order_uuid' => 'non_existent_order_uuid',
    ]);

    $this->assertInstanceOf(ResourceResponse::class, $response);
    $this->assertEquals(500, $response->getStatusCode());
    $this->assertEquals($this->t('Merchant Warrior: The order @order_id does not exist.', ['@order_id' => 'non_existent_order_uuid']), $response->getResponseData());
  }

  /**
   * Tests the POST method with an exception thrown during authorization.
   */
  public function testPostAuthorizationException(): void {
    // Set up the mocks for the entity storage.
    $this->setMocks();

    $this->apiClient->method('processAuthorization')
      ->will($this->throwException(new \Exception('Authorization failed')));

    $response = $this->resource->post([
      'merchant_warrior_access_token' => 'access_token',
      'merchant_warrior_payframe_token' => 'payframe_token',
      'merchant_warrior_payframe_key' => 'payframe_key',
      'order_uuid' => '1234',
    ]);

    $this->assertInstanceOf(ResourceResponse::class, $response);
    $this->assertEquals(500, $response->getStatusCode());
    $expected_response = [
      'status' => -1,
      'message' => 'Merchant Warrior: Payment authorization failed. The response received from the API is: <pre>@response</pre>',
    ];
    $this->assertEquals($expected_response, $response->getResponseData());
  }

  /**
   * Sets up the mocks for the entity storage.
   *
   * @param \Drupal\core\Entity\EntityStorageInterface|null $order_storage
   *   The order storage.
   * @param \Drupal\core\Entity\EntityStorageInterface|null $payment_gateway_storage
   *   The payment gateway storage.
   * @param \Drupal\core\Entity\EntityStorageInterface|null $payment_method_storage
   *   The payment method storage.
   * @param \Drupal\core\Entity\EntityStorageInterface|null $payment_storage
   *   The payment storage.
   */
  protected function setMocks($order_storage = NULL, $payment_gateway_storage = NULL, $payment_method_storage = NULL, $payment_storage = NULL) {
    if (!$order_storage) {
      // Create a mock for AddressItem.
      $address = new Address(
        'US', 'CA', 'Mountain View', 'MV', '94043', '94044',
        '1600 Amphitheatre Parkway', 'Google Bldg 41', 'Office 35',
        'Google Inc.', 'John', '', 'Smith', 'en'
      );

      $address_list = $this->prophesize(FieldItemListInterface::class);
      $address_list->first()->willReturn($address);
      $address_list = $address_list->reveal();

      // Create a mock for ProfileInterface.
      $billing_profile = $this->prophesize(ProfileInterface::class);
      $billing_profile->get('address')->willReturn($address_list);
      $billing_profile = $billing_profile->reveal();

      // Create a mock for OrderInterface.
      $order = $this->createMock(OrderInterface::class);
      $order->method('getTotalPrice')->willReturn(new Price('100.00', 'USD'));
      $order->method('getEmail')->willReturn('customer@example.com');
      $order->method('getBillingProfile')->willReturn($billing_profile);
      $order->method('uuid')->willReturn('1234');
      $order->method('id')->willReturn(1);
      // Mock the storage of 'commerce_order'.
      $order_storage = $this->createMock(EntityStorageInterface::class);
      $order_storage->expects($this->any())
        ->method('loadByProperties')
        ->with(['uuid' => '1234'])
        ->willReturn([$order]);
    }

    // Mock the storage for the payment entities.
    if (!$payment_method_storage) {
      $payment_method = $this->createMock(PaymentMethodInterface::class);
      $payment_method->method('save')->willReturn(TRUE);
      $payment_method->method('setExpiresTime')->willReturn($payment_method);
      $payment_method->method('setRemoteId')->willReturn($payment_method);
      $payment = $this->createMock(PaymentInterface::class);
      $payment->method('save')->willReturn(TRUE);
      $payment_method_storage = $this->createMock(EntityStorageInterface::class);
      $payment_method_storage->method('create')->willReturn($payment);
    }

    if (!$payment_storage) {
      $payment_storage = $this->createMock(EntityStorageInterface::class);
      $payment_storage->method('create')->willReturn($payment);
    }

    if (!$payment_gateway_storage) {
      $payment_gateway_plugin = $this->createMock(PaymentGatewayPaymentGatewayInterface::class);
      $payment_gateway_plugin->method('getMode')->willReturn('test');
      $payment_gateway = $this->createMock(PaymentGatewayInterface::class);
      $payment_gateway->method('id')->willReturn('merchant_warrior_payframe');
      $payment_gateway->method('getPlugin')->willReturn($payment_gateway_plugin);
      $payment_gateway_storage = $this->createMock(EntityStorageInterface::class);
      $payment_gateway_storage->method('load')->with('merchant_warrior_payframe')->willReturn($payment_gateway);
    }

    // Set up the entity type manager mock to return the storage mocks.
    $this->entityTypeManager->expects($this->any())
      ->method('getStorage')
      ->willReturnOnConsecutiveCalls(
        $order_storage,
        $payment_gateway_storage,
        $payment_method_storage,
        $payment_storage,
      );
  }

}
